/***************************************************************************
 *   Copyright (C) 2004 by Stefano Salvador                                *
 *   salva_ste@tin.it                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qcstring.h>
#include <qsocket.h>
#include <qdatetime.h>
#include <qbitarray.h>
#include <qfileinfo.h>

#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <kapplication.h>
#include <kdebug.h>
#include <kmessagebox.h>
#include <kinstance.h>
#include <kglobal.h>
#include <kstandarddirs.h>
#include <klocale.h>
#include <kurl.h>
#include <ksock.h>

#include "webshots.h"

using namespace KIO;

int BMP_START[] = {'B','M','6',0};
int BMP_END[] = {'B','M','F','+',0};
int JPEG_START[] = {0xff,0xd8,0xff,0xe0,0};
int JPEG_END[] = {0xff,0xd9,0x42,0};
int JPEG_END_P[] = {'B','M','F','+',0};
int ENCRYPT_P[] = {'W','W','B','B',0};

kio_webshotsProtocol::kio_webshotsProtocol(const QCString &pool_socket, const QCString &app_socket)
    : SlaveBase("kio_webshots", pool_socket, app_socket)
{
  kdDebug() << "kio_webshotsProtocol::kio_webshotsProtocol()" << endl;
  m_infoList = 0L;
  m_wsFile = 0L;
}


kio_webshotsProtocol::~kio_webshotsProtocol()
{
  kdDebug() << "kio_webshotsProtocol::~kio_webshotsProtocol()" << endl;
  delete m_infoList;
  delete m_wsFile;
}

bool kio_webshotsProtocol::checkNewFile( const KURL & url, QString & path )
{
  QString fullPath = url.path();
  
  // Are we already looking at that file ?
  if ( m_wsFile && m_wsName == fullPath.left(m_wsName.length()) )
  {
    // Has it changed ?
    struct stat statbuf;
    if ( ::stat( QFile::encodeName( m_wsName ), &statbuf ) == 0 )
    {
      if ( m_mtime == statbuf.st_mtime )
      {
        path = fullPath.mid( m_wsName.length() );
        return true;
      }
    }
  }

  // Close previous file
  if ( m_wsFile )
  {
    m_wsFile->close();
    delete m_wsFile;
    delete m_infoList;
    m_wsFile = 0L;
    m_infoList = 0L;
  }

  // Find where the file is in the full path
  int pos = 0;
  QString wsFile;
  path = QString::null;

  int len = fullPath.length();
  if ( len != 0 && fullPath[ len - 1 ] != '/' )
    fullPath += '/';

  while ( (pos=fullPath.find( '/', pos+1 )) != -1 )
  {
    QString tryPath = fullPath.left( pos );
    struct stat statbuf;
    if ( ::stat( QFile::encodeName(tryPath), &statbuf ) == 0 && !S_ISDIR(statbuf.st_mode) )
    {
      wsFile = tryPath;
      m_mtime = statbuf.st_mtime;
      path = fullPath.mid( pos + 1 );
      len = path.length();
      if ( len > 1 )
      {
        if ( path[ len - 1 ] == '/' )
          path.truncate( len - 1 );
      }
      else
        path = QString::fromLatin1("/");
      break;
    }
  }
  if ( wsFile.isEmpty() )
  {
    return false;
  }

  // Open new file
  m_wsFile = new QFile(wsFile);

  if ( !m_wsFile->open( IO_ReadOnly ) )
  {
    delete m_wsFile;
    delete m_infoList;
    m_wsFile = 0L;
    m_infoList = 0L;
    return false;
  }
  
  QFileInfo info(*m_wsFile);
  
  m_owner = info.owner();
  m_group = info.group();
  
  m_wsName = wsFile;
  return true;
}

void kio_webshotsProtocol::stat( const KURL & url )
{
  QString path;
  UDSEntry entry;
  
  if ( !checkNewFile( url, path ) )
  {
    // We may be looking at a real directory - this happens
    // when pressing up after being in the root of an archive
    QCString _path( QFile::encodeName(url.path()));
    struct stat buff;
    if ( ::stat( _path.data(), &buff ) == -1 || !S_ISDIR( buff.st_mode ) ) {
      error( KIO::ERR_DOES_NOT_EXIST, url.path() );
      return;
    }
    // Real directory. Return just enough information for KRun to work
    UDSAtom atom;
    atom.m_uds = KIO::UDS_NAME;
    atom.m_str = url.fileName();
    entry.append( atom );

    atom.m_uds = KIO::UDS_FILE_TYPE;
    atom.m_long = buff.st_mode & S_IFMT;
    entry.append( atom );

    statEntry( entry );
    
    finished();

    // And let go of the file - for people who want to unmount a cdrom after that
    delete m_wsFile;
    delete m_infoList;
    m_wsFile = 0L;
    m_infoList = 0L;
    return;
  }
  
  QStringList pathList = QStringList::split("/", path);
  if(pathList.count() < 2 && isCollection())
  {
    UDSAtom atom;
    entry.clear();
    atom.m_uds = UDS_FILE_TYPE;
    atom.m_long = S_IFDIR;
    entry.append( atom );
  }
  else
  {
    UDSAtom atom;
    entry.clear();
    atom.m_uds = UDS_FILE_TYPE;
    atom.m_long = S_IFREG;
    entry.append( atom );
    atom.m_uds = UDS_MIME_TYPE;
    atom.m_str = "image/x-wbz";
    entry.append( atom );
  }
  statEntry( entry );

  finished();
}

void kio_webshotsProtocol::listDir( const KURL & url )
{
  QString path;
  if ( !checkNewFile( url, path ) )
  {
    QCString _path( QFile::encodeName(url.path()));
    struct stat buff;
    if ( ::stat( _path.data(), &buff ) == -1 || !S_ISDIR( buff.st_mode ) ) {
      error( KIO::ERR_DOES_NOT_EXIST, url.prettyURL() );
      return;
    }
    // It's a real dir -> redirect
    KURL redir;
    redir.setPath( url.path() );
    redirection( redir );
    finished();
    // And let go of the tar file - for people who want to unmount a cdrom after that
    delete m_wsFile;
    delete m_infoList;
    m_wsFile = 0L;
    m_infoList = 0L;
    return;
  }

  QStringList pathList = QStringList::split("/", path);
  if(pathList.count() < 2)
  {
    UDSEntry entry;
    
    if(pathList.count() == 0)
    {
      UDSAtom atom;
      entry.clear();
      atom.m_uds = UDS_NAME;
      atom.m_str = getSectionName().latin1();
      entry.append(atom);
      atom.m_uds = UDS_FILE_TYPE;
      atom.m_long = S_IFDIR;
      entry.append( atom );
      atom.m_uds = UDS_MIME_TYPE;
      atom.m_str = "inode/directory";
      entry.append( atom );
      atom.m_uds = UDS_MODIFICATION_TIME;
      QFileInfo info(*m_wsFile);
      atom.m_long = info.lastModified().toTime_t();
      entry.append( atom );
      atom.m_uds = UDS_ACCESS;
      atom.m_long = 493;
      entry.append( atom );
      atom.m_uds = UDS_USER;
      atom.m_str = m_owner;
      entry.append( atom );
      atom.m_uds = UDS_GROUP;
      atom.m_str = m_group;
      entry.append( atom );
     
      listEntry( entry, false );    
    }
    else
    {
      getImagesList();
     
      totalSize( m_infoList->count() );
      QValueList<infoImg>::iterator it;
      for ( it = m_infoList->begin(); it != m_infoList->end(); it++ )
      {
        UDSAtom atom;
        entry.clear();
        atom.m_uds = UDS_NAME;
        atom.m_str = (*it).title.latin1();
        entry.append(atom);
        atom.m_uds = UDS_FILE_TYPE;
        atom.m_long = S_IFREG;
        entry.append( atom );
        atom.m_uds = UDS_MIME_TYPE;
        if((*it).type == 3)
          atom.m_str = "image/x-bmp";
        else if((*it).type == 2)
          atom.m_str = "image/jpeg";
        else
          atom.m_str = "image/x-wbz";
        entry.append( atom );
        atom.m_uds = UDS_MODIFICATION_TIME;
        QFileInfo info(*m_wsFile);
        atom.m_long = info.lastModified().toTime_t();
        entry.append( atom );
        atom.m_uds = UDS_SIZE;
        atom.m_long = (*it).size;
        entry.append( atom );
        atom.m_uds = UDS_ACCESS;
        atom.m_long = 292;
        entry.append( atom );
        atom.m_uds = UDS_USER;
        atom.m_str = m_owner;
        entry.append( atom );
        atom.m_uds = UDS_GROUP;
        atom.m_str = m_group;
        entry.append( atom );
  
        listEntry( entry, false );
      }
    }

    listEntry( entry, true );
  }

  finished();
}

void kio_webshotsProtocol::get(const KURL& url )
{
  QByteArray dataImg;
  
  QString path;
  if ( !checkNewFile( url, path ) )
  {
    QCString _path( QFile::encodeName(url.path()));
    struct stat buff;
    if ( ::stat( _path.data(), &buff ) == -1 || !S_ISDIR( buff.st_mode ) ) {
      error( KIO::ERR_DOES_NOT_EXIST, url.prettyURL() );
      return;
    }
    // It's a real dir -> redirect
    KURL redir;
    redir.setPath( url.path() );
    redirection( redir );
    finished();
    // And let go of the file - for people who want to unmount a cdrom after that
    delete m_wsFile;
    delete m_infoList;
    m_wsFile = 0L;
    m_infoList = 0L;
    return;
  }
  
  QStringList pathList = QStringList::split("/", path);
  if(pathList.size() == 2)
  {
    QString fileName = pathList[1];
    getImagesList();
    dataImg = getImage(fileName);
  }
  else if(!isCollection())
  {
    dataImg = getImage("");
  }
      
  // Send the data
  data(dataImg);
  data(QByteArray()); // empty array means we're done sending the data
  finished();
}


void kio_webshotsProtocol::mimetype(const KURL & /*url*/)
{
    mimeType("image/x-wbz");
    finished();
}

QString kio_webshotsProtocol::getSectionName()
{
  m_wsFile->at(0x000c);
  QString str;
  int c;
  while((c = m_wsFile->getch()) != 0){
    str.append(c);
  }
  return str;
}

void kio_webshotsProtocol::getImagesList()
{
  if(m_infoList)
    return;
  
  
  m_infoList = new QValueList<infoImg>;
  
  int pos = 0x0870;
  
  // Read images positions from header
  int i = 0; 
  while(1){
    int a,b,c,add;
    pos += 0x28;
    m_wsFile->at(pos);
    a = m_wsFile->getch();
    b = m_wsFile->getch();
    c = m_wsFile->getch();
    add = a + (b << 8) + (c << 16);
    
    if(add == 0)
      break;
      
    kdDebug() << "Loop " << add << endl;
    infoImg info(i++, add);
    m_infoList->append(info);
  }
  
  QValueList<infoImg>::iterator it;
  for ( it = m_infoList->begin(); it != m_infoList->end(); it++ )
  {
    int c;
    
    // Title
    m_wsFile->at((*it).address+0x010c);
    QString str;
    while((c = m_wsFile->getch()) != 0){
      str.append(c);
    }
    (*it).title = str;
    
    // Author
    str = "";
    m_wsFile->at((*it).address+0x028c);
    while((c = m_wsFile->getch()) != 0){
      str.append(c);
    }
    (*it).author = str;
    
    // Type
    str = "";
    m_wsFile->at((*it).address+0x038c);
    while((c = m_wsFile->getch()) != 0){
      str.append(c);
    }
    if( str == "wb0")
      (*it).type = 0;
    else if( str == "wb1")
      (*it).type = 1;
    else if( str == "jpg")
      (*it).type = 2;
    else if( str == "bmp")
      (*it).type = 3;
    else  
      (*it).type = -1;
  }
  
  for ( it = m_infoList->begin(); it != m_infoList->end(); it++ )
  {
    switch((*it).type)
    {
      case 0:  // JPEG WB0 Encrypted
      case 1:  // JPEG WB1 Encrypted
        (*it).end = search((*it).address, m_wsFile->size(), JPEG_END_P);  
        (*it).start = search((*it).address, (*it).end, ENCRYPT_P);
        break;
      case 2:  // Plain JPEG
        (*it).end = search((*it).address, m_wsFile->size(), JPEG_END);
        (*it).start = search((*it).address, (*it).end, JPEG_START) - 8; // see the + 4 in decriptImage
        break;
      case 3:  // Plain BMP
        (*it).end = search((*it).address, m_wsFile->size(), BMP_END);
        (*it).start = search((*it).address, (*it).end, BMP_START) - 7; // see the + 4 in decriptImage
        break;
      default:
        (*it).isValid = false;
        return;
    }
  
    (*it).size = (*it).end - (*it).start + 4;
    (*it).isValid = (*it).size > 5;
  } 
}

QByteArray kio_webshotsProtocol::getImage(QString fileName)
{ 
  QIODevice::Offset start;
  QIODevice::Offset size;
  int type;
  
  if(!fileName.isEmpty())
  {
    QValueList<infoImg>::iterator it;
    for ( it = m_infoList->begin(); it != m_infoList->end(); it++ )
    {
      if((*it).title == fileName)
        break;
    }
  
    start = (*it).start;
    size = (*it).size;
    type = (*it).type;
  }
  else
  {
    QIODevice::Offset end = m_wsFile->size();
    start = search(0, end, ENCRYPT_P);
    size = end - start + 4;
    m_wsFile->at(start + 1);
    m_wsFile->getch();
    type = m_wsFile->getch() - 0x30;
  }
  
  if(type == 3)
    mimeType("image/x-bmp");
  else if(type == 2)
    mimeType("image/jpeg");
  else
  {
    mimeType("image/x-wbz");
    start -= 14; // to add wbz identifier
    size += 14;
  }

  m_wsFile->at(start + 4);
  QByteArray theImage(size);
  char * enc = theImage.data();
  m_wsFile->readBlock(enc, size);
  
  if(type == 0 || type == 1)
  { 
    // adding wbz header
    theImage[0] = 0x43;
    theImage[1] = 0xab;
    theImage[2] = 0x91;
    theImage[3] = 0x67;
    theImage[4] = 0x98;
    theImage[5] = 0x00;
  }
    
  return theImage;  
}

int kio_webshotsProtocol::search(unsigned int from, unsigned int to, int str[])
{
  m_wsFile->at(from);
  while(m_wsFile->at() < to)
  {
    int i = 0;
    while(str[i] != 0)
    {
      if( m_wsFile->getch() != str[i++]){
        goto next;
      }
    }
    
    return m_wsFile->at();
    
    next:
      continue;
  }
  return 0;
}

bool kio_webshotsProtocol::isCollection()
{
  int WBZ_ID[] = {0x43,0xab,0x91,0x67,0x98,0};
  
  m_wsFile->at(0);
  for(int i=0; i<6; i++)
  {
    if(m_wsFile->getch() != WBZ_ID[i])
      return true;
  }
  return false;
}

extern "C"
{
    int kdemain(int argc, char **argv)
    {
        KInstance instance( "kio_webshots" );
        
        kdDebug(7101) << "*** Starting kio_webshots " << endl;
        
        if (argc != 4) {
            kdDebug(7101) << "Usage: kio_webshots  protocol domain-socket1 domain-socket2" << endl;
            exit(-1);
        }
        
        kio_webshotsProtocol slave(argv[2], argv[3]);
        slave.dispatchLoop();
        
        kdDebug(7101) << "*** kio_webshots Done" << endl;
        return 0;
    }
} 
